@@ -46,11 +46,19 @@
 
 #include <arpa/inet.h>
 #include <net/if.h>
-#include <netinet/if_ether.h>
 #include <netinet/in_systm.h>
 #include <netinet/in.h>
 #include <net/route.h>
 
+/* musl has its own definition of struct ethhdr, so only include
+ * netinet/if_ether.h on systems with GLIBC.  For the ARPHRD constants,
+ * we must include linux/if_arp.h instead. */
+#if defined(__GLIBC__)
+#include <netinet/if_ether.h>
+#else
+#include <linux/if_arp.h>
+#endif
+
 #include <errno.h>
 #include <fcntl.h>
 #include <ctype.h>
@@ -117,6 +125,8 @@ static const uint8_t ipv4_bcast_addr[] = {
 };
 #endif
 
+static int if_addressexists(struct interface *, struct in_addr *);
+
 #define PROC_INET6	"/proc/net/if_inet6"
 #define PROC_PROMOTE	"/proc/sys/net/ipv4/conf/%s/promote_secondaries"
 #define SYS_BRIDGE	"/sys/class/net/%s/bridge"
@@ -219,7 +229,7 @@ check_proc_hex(const char *path, unsigned int *value)
 }
 
 static ssize_t
-write_path(const char *path, const char *val)
+if_writepathuint(const char *path, unsigned int val)
 {
 	FILE *fp;
 	ssize_t r;
@@ -227,7 +237,7 @@ write_path(const char *path, const char *val)
 	fp = fopen(path, "w");
 	if (fp == NULL)
 		return -1;
-	r = fprintf(fp, "%s\n", val);
+	r = fprintf(fp, "%u\n", val);
 	fclose(fp);
 	return r;
 }
@@ -251,7 +261,7 @@ if_init(struct interface *ifp)
 		return errno == ENOENT ? 0 : -1;
 	if (n == 1)
 		return 0;
-	return write_path(path, "1") == -1 ? -1 : 0;
+	return if_writepathuint(path, 1) == -1 ? -1 : 0;
 }
 
 int
@@ -325,7 +335,7 @@ if_vlanid(const struct interface *ifp)
 }
 
 static int
-_open_link_socket(struct sockaddr_nl *nl, int protocol)
+if_linksocket(struct sockaddr_nl *nl, int protocol)
 {
 	int fd;
 
@@ -362,7 +372,7 @@ if_opensockets_os(struct dhcpcd_ctx *ctx)
 	snl.nl_groups |= RTMGRP_IPV6_ROUTE | RTMGRP_IPV6_IFADDR | RTMGRP_NEIGH;
 #endif
 
-	ctx->link_fd = _open_link_socket(&snl, NETLINK_ROUTE);
+	ctx->link_fd = if_linksocket(&snl, NETLINK_ROUTE);
 	if (ctx->link_fd == -1)
 		return -1;
 #ifdef NETLINK_BROADCAST_ERROR
@@ -376,7 +386,7 @@ if_opensockets_os(struct dhcpcd_ctx *ctx)
 
 	ctx->priv = priv;
 	memset(&snl, 0, sizeof(snl));
-	priv->route_fd = _open_link_socket(&snl, NETLINK_ROUTE);
+	priv->route_fd = if_linksocket(&snl, NETLINK_ROUTE);
 	if (priv->route_fd == -1)
 		return -1;
 	len = sizeof(snl);
@@ -407,9 +417,8 @@ if_carrier(struct interface *ifp)
 }
 
 static int
-get_netlink(struct dhcpcd_ctx *ctx, struct iovec *iov,
-    void *arg, int fd, int flags,
-    int (*callback)(struct dhcpcd_ctx *, void *, struct nlmsghdr *))
+if_getnetlink(struct dhcpcd_ctx *ctx, struct iovec *iov, int fd, int flags,
+    int (*cb)(struct dhcpcd_ctx *, void *, struct nlmsghdr *), void *cbarg)
 {
 	struct sockaddr_nl nladdr = { .nl_pid = 0 };
 	struct msghdr msg = {
@@ -437,7 +446,7 @@ recv_again:
 		return 0;
 
 	r = 0;
-	again = 0;	/* Appease static analysis */
+	again = 0;
 	for (nlm = iov->iov_base;
 	     nlm && NLMSG_OK(nlm, (size_t)len);
 	     nlm = NLMSG_NEXT(nlm, len))
@@ -463,7 +472,7 @@ recv_again:
 			again = 0;
 			break;
 		}
-		if (callback && (r = callback(ctx, arg, nlm)) != 0)
+		if (cb != NULL && (r = cb(ctx, cbarg, nlm)) != 0)
 			break;
 	}
 
@@ -626,6 +635,7 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 #endif
 #ifdef INET6
 	struct in6_addr addr6;
+	int flags;
 #endif
 
 	if (nlm->nlmsg_type != RTM_DELADDR && nlm->nlmsg_type != RTM_NEWADDR)
@@ -674,6 +684,17 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 			}
 			rta = RTA_NEXT(rta, len);
 		}
+
+		/* Validate RTM_DELADDR really means address deleted
+		 * and anything else really means address exists. */
+		if (if_addressexists(ifp, &addr) == 1) {
+			if (nlm->nlmsg_type == RTM_DELADDR)
+				break;
+		} else {
+			if (nlm->nlmsg_type != RTM_DELADDR)
+				break;
+		}
+
 		ipv4_handleifa(ctx, nlm->nlmsg_type, NULL, ifp->name,
 		    &addr, &net, &brd, ifa->ifa_flags, (pid_t)nlm->nlmsg_pid);
 		break;
@@ -690,6 +711,18 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 			}
 			rta = RTA_NEXT(rta, len);
 		}
+
+		/* Validate RTM_DELADDR really means address deleted
+		 * and anything else really means address exists. */
+		flags = if_addrflags6(ifp, &addr6, NULL);
+		if (nlm->nlmsg_type == RTM_DELADDR) {
+			if (flags != -1)
+				break;
+		} else {
+			if (flags == -1)
+				break;
+		}
+
 		ipv6_handleifa(ctx, nlm->nlmsg_type, NULL, ifp->name,
 		    &addr6, ifa->ifa_prefixlen, ifa->ifa_flags,
 		    (pid_t)nlm->nlmsg_pid);
@@ -873,14 +906,13 @@ if_handlelink(struct dhcpcd_ctx *ctx)
 		.iov_len = sizeof(buf),
 	};
 
-	return get_netlink(ctx, &iov, NULL,
-	    ctx->link_fd, MSG_DONTWAIT, &link_netlink);
+	return if_getnetlink(ctx, &iov, ctx->link_fd, MSG_DONTWAIT,
+	    &link_netlink, NULL);
 }
 
 static int
-send_netlink(struct dhcpcd_ctx *ctx, void *arg,
-    int protocol, struct nlmsghdr *hdr,
-    int (*callback)(struct dhcpcd_ctx *, void *, struct nlmsghdr *))
+if_sendnetlink(struct dhcpcd_ctx *ctx, int protocol, struct nlmsghdr *hdr,
+    int (*cb)(struct dhcpcd_ctx *, void *, struct nlmsghdr *), void *cbarg)
 {
 	int s, r;
 	struct sockaddr_nl snl = { .nl_family = AF_NETLINK };
@@ -890,15 +922,18 @@ send_netlink(struct dhcpcd_ctx *ctx, void *arg,
 	    .msg_iov = &iov, .msg_iovlen = 1
 	};
 
-	if (protocol == NETLINK_ROUTE) {
-		struct priv *priv;
+	if ((s = if_linksocket(&snl, protocol)) == -1)
+		return -1;
 
-		priv = (struct priv *)ctx->priv;
-		s = priv->route_fd;
-	} else {
-		if ((s = _open_link_socket(&snl, protocol)) == -1)
-			return -1;
+#ifdef NETLINK_GET_STRICT_CHK
+	if (hdr->nlmsg_type == RTM_GETADDR) {
+		int on = 1;
+
+		if (setsockopt(s, SOL_NETLINK, NETLINK_GET_STRICT_CHK,
+		    &on, sizeof(on)) == -1)
+			logerr("%s: NETLINK_GET_STRICT_CHK", __func__);
 	}
+#endif
 
 	/* Request a reply */
 	hdr->nlmsg_flags |= NLM_F_ACK;
@@ -910,11 +945,10 @@ send_netlink(struct dhcpcd_ctx *ctx, void *arg,
 			.iov_len = sizeof(buf),
 		};
 
-		r = get_netlink(ctx, &riov, arg, s, 0, callback);
+		r = if_getnetlink(ctx, &riov, s, 0, cb, cbarg);
 	} else
 		r = -1;
-	if (protocol != NETLINK_ROUTE)
-		close(s);
+	close(s);
 	return r;
 }
 
@@ -1117,8 +1151,8 @@ gnl_getfamily(struct dhcpcd_ctx *ctx, const char *name)
 	if (nla_put_string(&nlm.hdr, sizeof(nlm),
 	    CTRL_ATTR_FAMILY_NAME, name) == -1)
 		return -1;
-	return send_netlink(ctx, NULL, NETLINK_GENERIC, &nlm.hdr,
-	    &_gnl_getfamily);
+	return if_sendnetlink(ctx, NETLINK_GENERIC, &nlm.hdr,
+	    &_gnl_getfamily, NULL);
 }
 
 static int
@@ -1193,7 +1227,8 @@ if_getssid_nl80211(struct interface *ifp)
 	nlm.hdr.nlmsg_flags = NLM_F_REQUEST;
 	nlm.ghdr.cmd = NL80211_CMD_GET_WIPHY;
 	nla_put_32(&nlm.hdr, sizeof(nlm), NL80211_ATTR_IFINDEX, ifp->index);
-	if (send_netlink(ifp->ctx, ifp, NETLINK_GENERIC, &nlm.hdr, NULL) == -1)
+	if (if_sendnetlink(ifp->ctx, NETLINK_GENERIC, &nlm.hdr,
+	    NULL, NULL) == -1)
 		return -1;
 
 	/* We need to parse out the list of scan results and find the one
@@ -1205,8 +1240,8 @@ if_getssid_nl80211(struct interface *ifp)
 	nlm.ghdr.cmd = NL80211_CMD_GET_SCAN;
 	nla_put_32(&nlm.hdr, sizeof(nlm), NL80211_ATTR_IFINDEX, ifp->index);
 
-	return send_netlink(ifp->ctx, ifp,
-	    NETLINK_GENERIC, &nlm.hdr, &_if_getssid_nl80211);
+	return if_sendnetlink(ifp->ctx, NETLINK_GENERIC, &nlm.hdr,
+	    &_if_getssid_nl80211, ifp);
 }
 #endif
 
@@ -1236,6 +1271,60 @@ struct nlma
 	char buffer[64];
 };
 
+#ifdef INET
+struct ifiaddr
+{
+	unsigned int ifa_ifindex;
+	struct in_addr ifa_addr;
+};
+
+static int
+_if_addressexists(__unused struct dhcpcd_ctx *ctx,
+    void *arg, struct nlmsghdr *nlm)
+{
+	struct ifiaddr *ia = arg;
+	in_addr_t this_addr;
+	size_t len;
+	struct rtattr *rta;
+	struct ifaddrmsg *ifa;
+
+	ifa = NLMSG_DATA(nlm);
+	if (ifa->ifa_index != ia->ifa_ifindex || ifa->ifa_family != AF_INET)
+		return 0;
+	rta = (struct rtattr *)IFA_RTA(ifa);
+	len = NLMSG_PAYLOAD(nlm, sizeof(*ifa));
+	while (RTA_OK(rta, len)) {
+		switch (rta->rta_type) {
+		case IFA_LOCAL:
+			memcpy(&this_addr, RTA_DATA(rta), sizeof(this_addr));
+			return this_addr == ia->ifa_addr.s_addr ? 1 : 0;
+		}
+		rta = RTA_NEXT(rta, len);
+	}
+	return 0;
+}
+
+static int
+if_addressexists(struct interface *ifp, struct in_addr *addr)
+{
+	struct ifiaddr ia = {
+		.ifa_ifindex = ifp->index,
+		.ifa_addr = *addr,
+	};
+	struct nlma nlm = {
+	    .hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifaddrmsg)),
+	    .hdr.nlmsg_type = RTM_GETADDR,
+	    .hdr.nlmsg_flags = NLM_F_DUMP | NLM_F_REQUEST,
+	    .ifa.ifa_family = AF_INET,
+	    .ifa.ifa_index = ifp->index,
+	};
+
+	return if_sendnetlink(ifp->ctx, NETLINK_ROUTE, &nlm.hdr,
+	    &_if_addressexists, &ia);
+}
+#endif
+
+
 struct nlmr
 {
 	struct nlmsghdr hdr;
@@ -1342,8 +1431,8 @@ if_route(unsigned char cmd, const struct rt *rt)
 		add_attr_32(&nlm.hdr, sizeof(nlm), RTA_PRIORITY,
 		    rt->rt_metric);
 
-	return send_netlink(rt->rt_ifp->ctx, NULL,
-	    NETLINK_ROUTE, &nlm.hdr, NULL);
+	return if_sendnetlink(rt->rt_ifp->ctx, NETLINK_ROUTE, &nlm.hdr,
+	    NULL, NULL);
 }
 
 static int
@@ -1376,7 +1465,8 @@ if_initrt(struct dhcpcd_ctx *ctx, rb_tree_t *kroutes, int af)
 	    .rt.rtm_family = (unsigned char)af,
 	};
 
-	return send_netlink(ctx, kroutes, NETLINK_ROUTE, &nlm.hdr, &_if_initrt);
+	return if_sendnetlink(ctx, NETLINK_ROUTE, &nlm.hdr,
+	    &_if_initrt, kroutes);
 }
 
 
@@ -1574,8 +1664,8 @@ if_address(unsigned char cmd, const struct ipv4_addr *ia)
 		    &cinfo, sizeof(cinfo));
 	}
 
-	if (send_netlink(ia->iface->ctx, NULL,
-	    NETLINK_ROUTE, &nlm.hdr, NULL) == -1)
+	if (if_sendnetlink(ia->iface->ctx, NETLINK_ROUTE, &nlm.hdr,
+	    NULL, NULL) == -1)
 		retval = -1;
 	return retval;
 }
@@ -1649,8 +1739,8 @@ if_address6(unsigned char cmd, const struct ipv6_addr *ia)
 		    &cinfo, sizeof(cinfo));
 	}
 
-	return send_netlink(ia->iface->ctx, NULL,
-	    NETLINK_ROUTE, &nlm.hdr, NULL);
+	return if_sendnetlink(ia->iface->ctx, NETLINK_ROUTE, &nlm.hdr,
+	    NULL, NULL);
 }
 
 int
@@ -1754,7 +1844,7 @@ if_disable_autolinklocal(struct dhcpcd_ctx *ctx, unsigned int ifindex)
 	add_attr_nest_end(&nlm.hdr, afs6);
 	add_attr_nest_end(&nlm.hdr, afs);
 
-	return send_netlink(ctx, NULL, NETLINK_ROUTE, &nlm.hdr, NULL);
+	return if_sendnetlink(ctx, NETLINK_ROUTE, &nlm.hdr, NULL, NULL);
 #else
 	UNUSED(ctx);
 	UNUSED(ifindex);
@@ -1763,7 +1853,8 @@ if_disable_autolinklocal(struct dhcpcd_ctx *ctx, unsigned int ifindex)
 #endif
 }
 
-static const char *prefix = "/proc/sys/net/ipv6/conf";
+static const char *p_conf = "/proc/sys/net/ipv6/conf";
+static const char *p_neigh = "/proc/sys/net/ipv6/neigh";
 
 void
 if_setup_inet6(const struct interface *ifp)
@@ -1782,14 +1873,14 @@ if_setup_inet6(const struct interface *ifp)
 	if (!(ifp->options->options & DHCPCD_IPV6RS))
 		return;
 
-	snprintf(path, sizeof(path), "%s/%s/autoconf", prefix, ifp->name);
+	snprintf(path, sizeof(path), "%s/%s/autoconf", p_conf, ifp->name);
 	ra = check_proc_int(path);
 	if (ra != 1 && ra != -1) {
-		if (write_path(path, "0") == -1)
+		if (if_writepathuint(path, 0) == -1)
 			logerr("%s: %s", __func__, path);
 	}
 
-	snprintf(path, sizeof(path), "%s/%s/accept_ra", prefix, ifp->name);
+	snprintf(path, sizeof(path), "%s/%s/accept_ra", p_conf, ifp->name);
 	ra = check_proc_int(path);
 	if (ra == -1) {
 		logfunc_t *logfunc = errno == ENOENT? logdebug : logwarn;
@@ -1798,11 +1889,31 @@ if_setup_inet6(const struct interface *ifp)
 		 * error as such so just log it and continue */
 		logfunc("%s", path);
 	} else if (ra != 0) {
-		if (write_path(path, "0") == -1)
+		if (if_writepathuint(path, 0) == -1)
 			logerr("%s: %s", __func__, path);
 	}
 }
 
+int
+if_applyra(const struct ra *rap)
+{
+	char path[256];
+	const char *ifname = rap->iface->name;
+	int error = 0;
+
+	snprintf(path, sizeof(path), "%s/%s/hop_limit", p_conf, ifname);
+	if (if_writepathuint(path, rap->hoplimit) == -1)
+		error = -1;
+	snprintf(path, sizeof(path), "%s/%s/retrans_time", p_neigh, ifname);
+	if (if_writepathuint(path, rap->retrans) == -1)
+		error = -1;
+	snprintf(path, sizeof(path), "%s/%s/base_reachable_time",
+	    p_neigh, ifname);
+	if (if_writepathuint(path, rap->reachable) == -1)
+		error = -1;
+	return error;
+}
+
 #ifdef IPV6_MANAGETEMPADDR
 int
 ip6_use_tempaddr(const char *ifname)
@@ -1812,7 +1923,7 @@ ip6_use_tempaddr(const char *ifname)
 
 	if (ifname == NULL)
 		ifname = "all";
-	snprintf(path, sizeof(path), "%s/%s/use_tempaddr", prefix, ifname);
+	snprintf(path, sizeof(path), "%s/%s/use_tempaddr", p_conf, ifname);
 	val = check_proc_int(path);
 	return val == -1 ? 0 : val;
 }
@@ -1825,7 +1936,7 @@ ip6_temp_preferred_lifetime(const char *ifname)
 
 	if (ifname == NULL)
 		ifname = "all";
-	snprintf(path, sizeof(path), "%s/%s/temp_prefered_lft", prefix,
+	snprintf(path, sizeof(path), "%s/%s/temp_prefered_lft", p_conf,
 	    ifname);
 	val = check_proc_int(path);
 	return val < 0 ? TEMP_PREFERRED_LIFETIME : val;
@@ -1839,7 +1950,7 @@ ip6_temp_valid_lifetime(const char *ifname)
 
 	if (ifname == NULL)
 		ifname = "all";
-	snprintf(path, sizeof(path), "%s/%s/temp_valid_lft", prefix, ifname);
+	snprintf(path, sizeof(path), "%s/%s/temp_valid_lft", p_conf, ifname);
 	val = check_proc_int(path);
 	return val < 0 ? TEMP_VALID_LIFETIME : val;
 }
@@ -1853,7 +1964,7 @@ ip6_forwarding(const char *ifname)
 
 	if (ifname == NULL)
 		ifname = "all";
-	snprintf(path, sizeof(path), "%s/%s/forwarding", prefix, ifname);
+	snprintf(path, sizeof(path), "%s/%s/forwarding", p_conf, ifname);
 	val = check_proc_int(path);
 	return val == -1 ? 0 : val;
 }
