@@ -1711,6 +1711,19 @@ if_getlifetime6(struct ipv6_addr *addr)
 	return -1;
 }
 
+int
+if_applyra(const struct ra *rap)
+{
+	struct lifreq		lifr = {
+		.lifr_ifinfo.lir_maxhops = rap->hoplimit,
+		.lifr_ifinfo.lir_reachtime = rap->reachable,
+		.lifr_ifinfo.lir_reachretrans = rap->retrans,
+	};
+
+	strlcpy(lifr.lifr_name, rap->iface->name, sizeof(lifr.lifr_name));
+	return ioctl(rap->iface->ctx->pf_inet_fd, SIOCSLIFLNKINFO, &lifr);
+}
+
 void
 if_setup_inet6(__unused const struct interface *ifp)
 {
